import time
from threading import Timer as _Timer

from .util import maybe_lock
    
class Stopwatch:
    def __init__(self, punch:bool=True):
        self.t = None
        if punch:
            self.punch()

    def punch(self, latency:float=0):
        """punch the clock and return elapsed time since previous punch
        
        Args:
            latency: punch `latency` seconds in the past, 
                unless it would be before the previous punch
        """
        t = time.perf_counter_ns() - latency
        if self.t is None:
            dt_ns = 0
        else:
            t = max(self.t, t)
            dt_ns = t - self.t
        self.t = t
        return dt_ns * 1e-9

    def read(self):
        """just return elapsed time since last punch"""
        if self.t is None:
            return self.punch()
        return (time.perf_counter_ns() - self.t) * 1e-9
    
class Timer:
    """a threading.Timer using the global iipyper lock around the timed function.
    also starts automatically by default.
    """
    def __init__(self, interval, f, lock=True, start=True, **kw):
        self.timer = _Timer(max(0,interval), maybe_lock(f, lock), **kw)
        if start:
            self.start()
    def cancel(self):
        self.timer.cancel()
    def start(self):
        self.timer.start()


import time
import json
import pickle
import inspect
import typing
from threading import Thread

# from pythonosc import osc_packet
# from pythonosc.osc_server import AsyncIOOSCUDPServer
from pythonosc.osc_server import BlockingOSCUDPServer, ThreadingOSCUDPServer
from pythonosc.dispatcher import Dispatcher
from pythonosc.udp_client import SimpleUDPClient

import pydantic

from .types import *
from .util import maybe_lock

# leaving this here for now. seems like it may not be useful since nested bundles
# do not appear to work in sclang.
# class BundleDispatcher(Dispatcher):
#     def call_handlers_for_packet(self,
#              data: bytes, client_address: Tuple[str, int]
#              ) -> None:
#         """Override python-osc to handle whole bundles in one callback.
#             data: Data of packet
#             client_address: Address of client this packet originated from
#         """
#         # Get OSC messages from all bundles or standalone message.
#         try:
#             packet = osc_packet.OscPacket(data)
#             addrs = [msg.message.address for msg in packet.messages]
#             # get common root address
#             root_addr = '/root' # TODO
#             stem_addrs = addrs # TODO
#             handler = self.handlers_for_address(root_addr)
#             # TODO: handle time
#             # times = [msg.time for msg in packet.messages]
#             fused_message = []
#             for stem, msg in zip(stem_addrs, packet.messages):
#                 fused_message.append(stem)
#                 fused_message.append(msg.message)
#             handler.invoke(client_address, fused_message)
#         except osc_packet.ParseError:
#             pass

# can't use None for this since and argument value might be None
class _Eos:
    pass
_eos = _Eos()

def _consume_splat(hd:Any, tl:Iterable, cls:type, is_key) -> Tuple[Any, Any]:
    # get length of Splat from type parameter
    params = typing.get_args(cls.__supertype__)
    splat_items = []

    if len(params)==0:
        # print(cls)
        # Splat[None] case
        # read until a string or end of iteration is encountered
        while hd is not _eos and not is_key(hd):
            splat_items.append(hd)
            # print(f'{splat_items=}')
            hd = next(tl, _eos)
        return splat_items, hd
    
    else:
        # print(cls)
        # Splat[N] case
        # read the annotated number of items
        for _ in range(len(params)):
            if hd is _eos:
                raise ValueError(f"""
                hit end of arguments while parsing {cls}
                """)
            splat_items.append(hd)
            # print(f'{splat_items=}')
            hd = next(tl, _eos)
        return splat_items, hd
    
def _is_legacy_json_str(item):
    if isinstance(item, str) and item.startswith('%JSON:'):
        return item[6:], True
    else:
        return item, False

def _consume_items(hd:Any, tl:Iterable, cls:type, is_key) -> Tuple[Any, Any]:
    """
    Args:
        hd: first element of items
        tl: remaining items (an iterator)
        cls: annotated type to consume
        is_key: test if an item indicates an argument name
    Returns:
        instance of cls constructed from consumed items,
        next item following those consumed (or None if end of iteration)
    """
    # print(f'{hd=}, {tl=}, {cls=}')

    # consume groups of items annotated as Vector
    if hasattr(cls, '__name__') and cls.__name__=='Splat':
        return _consume_splat(hd, tl, cls, is_key)
    
    hd, is_legacy_json = _is_legacy_json_str(hd)

    if cls is object or is_legacy_json:
        # interpret object-anotated strings as JSON
        if isinstance(hd, str):
            hd = json.loads(hd)
        # interpret object-annotated blobs as pickled objects
        elif isinstance(hd, bytes):
            # pickle
            hd = pickle.loads(hd)

    if cls is NDArray:
        # interpret NDArrray-anotated strings as iipyper JSON or numpy repr format
        if isinstance(hd, str):
            if hd.startswith('array'):
                hd = ndarray_from_repr(hd)
            else:
                hd = ndarray_from_json(hd)
        # interpret NDArray-annotated blobs as raw float32 buffers
        elif isinstance(hd, bytes):
            hd = np.frombuffer(hd, dtype=np.float32)

    #single item case: return the head, advance to first element of tail
    return hd, next(tl, _eos)

def _parse_osc_items(osc_items, sig_info, allow_pos, allow_kw, verbose) -> Tuple[List, Dict[str, Any]]:
    """
    convert a list of osc message contents to positional and keyword arguments
    """
    positional_params, named_params, has_varp, has_varkw = sig_info

    if verbose > 1:
        print(f"""parsing OSC
        {osc_items=}
        {positional_params=}
        {named_params=}
        {has_varp=} 
        {has_varkw=}""")

    position = 0 if len(positional_params) or has_varp else -1

    items = iter(osc_items)
    args = []
    kw = {}

    # decide if an item represents the name of an argument,
    # or is a positional argument or part of a Splat[None]
    def is_key(item):
        if verbose > 2:
            print(f"""\tparsing {item=}
                {allow_kw=} 
                and {isinstance(item, str)=} 
                and ({(item in named_params)=} 
                    or ({has_varkw=} and {(position<0)=}) or not {allow_pos=}))""")
        return (
            allow_kw # named arguments enabled
            and isinstance(item, str) # can be a name
            and (item in named_params # is a known name
                or (has_varkw and position<0) or not allow_pos) # or must be a ** arg
        )
    try:
        _, item = _consume_items(None, items, None, is_key)
        while item is not _eos:
            # print(item, is_key(item))
            if is_key(item):
                ### interpret this item as the name of an argument
                key = item
                # print(f'{key=}')
                if item in named_params:
                    cls = named_params[item].annotation
                else:
                    cls = Any
                try:
                    value, item = _consume_items(
                        next(items), items, cls, is_key)
                except json.JSONDecodeError:
                    print(f'JSON error decoding argument "{key}"')
                    raise
                kw[key] = value
                position = -1 # no more positional arguments allowed
            elif position>=0:
                ### interpret this item as a positional argument
                # print(f'{position=}')
                if position < len(positional_params):
                    cls = positional_params[position].annotation
                elif has_varp:
                    cls = Any
                else:
                    raise ValueError("""
                    too many positional arguments.
                    """)
                # print(f'{cls=}')
                # print(position, positional_params)
                if (
                    cls is Splat[None] 
                    and not allow_kw 
                    and position<len(positional_params)-1
                    ):
                    raise ValueError("""
                    Splat[None] can only be used on the last argument when kwargs=False.
                    """)
                try:
                    value, item = _consume_items(
                        item, items, cls, is_key)
                except json.JSONDecodeError:
                    print('JSON error decoding argument', position)
                    raise
                args.append(value)
                position += 1
                # check if this was the final positional argument
                if position >= len(positional_params) and not has_varp:
                    position = -1
            else:
                raise ValueError(f"""
                positional argument after keyword arg while parsing OSC.
                parsing: {osc_items}
                positional arguments parsed: {args}
                keyword arguments parsed: {kw}
                failed on item: {item}
                """)
    except StopIteration:
        pass

    if verbose > 2:
        print(f"""parsed
            {args=} 
            {kw=}""")
    return args, kw


class OSC():
    """
    iipyper OSC object.
    Create one of these and use it to make OSC handlers and send OSC:

    ```python
    # make an OSC object receiving on port 9999
    osc = OSC(port=9999)

    # receive OSC messages at '/my_route'
    @osc.handle
    def my_route(route, *osc_items):
        print(route, osc_items)

    # with wildcard route and replies:
    @osc.handle('/my_wildcard_route/*')
    def _(route, *osc_items):
        print(route, osc_items)
        # return value sends a reply
        return '/echo' + route, *osc_items

    # create an OSC client and send it a message
    osc.create_client('my_client', host='127.0.0.1', port=8888)
    osc.send('/my/osc/route', 0, 1.0, 'abc', client='my_client')
    ```
    """
    def __init__(self, 
        host:str="127.0.0.1", port:int=9999, 
        verbose:int=1, concurrent:bool=False):
        """
        TODO: Expand to support multiple IPs + ports

        Args:
            host (str): IP address
            port (int): port to receive on
            verbose (bool): whether to print activity
            concurrent (bool): if True, handle each incoming OSC message on 
                its own thread. otherwise, incoming OSC is handled serially on 
                one thread for the whole OSC object.
        """
        self.verbose = verbose
        self.concurrent = concurrent
        self.host = host
        self.port = port
        self.dispatcher = Dispatcher()
        self.server = None
        self.clients = {} # (host,port) -> client
        self.client_names = {} # (name) -> (host,port)

        self.create_server()

    def create_server(self):#, host=None, port=None):
        """
        Create the server
        """
        # if (host is None):
        #     host = self.host
        # if (port is None):
        #     port = self.port
        cls = ThreadingOSCUDPServer if self.concurrent else BlockingOSCUDPServer

        if (self.server is None):
            self.server = cls((self.host, self.port), self.dispatcher)
            if self.verbose > 0:
                print(f"OSC server created {self.host}:{self.port}")

            # start the OSC server on its own thread
            Thread(target=self.server.serve_forever, daemon=True).start()
            # self.server.serve_forever()
        else:
            print("OSC server already exists")

    # def close_server(self):
    #     """
    #     Close the server
    #     """
    #     if (self.server is not None):
    #         self.transport.close()
    #     else:
    #         print("OSC server does not exist")

    def add_handler(self, address, handler):
        """
        Map the custom message handler to the OSC dispatcher
        """
        # if (self.server is not None):
        self.dispatcher.map(address, handler, needs_reply_address=True)

    def create_client(self, name:str, host:Optional[str]=None, port:Optional[int]=None):
        """
        Add an OSC client.
        Args:
            name: name this client
            host (str): IP to send to, defaults to same as server
            port (int): port to send to, defaults to 57120 (supercollider)
        """
        if (host == None):
            host = self.host
        if (port == None):
            port = 57120
        if ((host, port) not in self.clients):
            self.clients[host, port] = SimpleUDPClient(host, port)
            if self.verbose > 0:
                print(f"OSC client created {host}:{port}")
        else:
            print("OSC client already exists")
        self.client_names[name] = (host, port)

    def get_client_by_name(self, name):
        try:
            return self.clients[self.client_names[name]]
        except Exception:
            print(f'no client with name "{name}"')
            return None

    def get_client_by_sender(self, address):
        if address not in self.clients:
            host, port = address
            self.create_client(f'{host}:{port}', host, port)
        return self.clients[address]

    def send(self, route:str, *msg, client:Optional[str]=None):
        """
        Send message to default client, or with client in address

        Args:
            route: e.g. '/my/osc/route' or 'host:port/my/osc/route'
            *msg: contents of OSC message
            client: name of OSC client or None to use default client
        """
        if len(self.clients)==0:
            print('ERROR: iipyper: send: no OSC clients. use `create_client` to make one.')
            return

        if client is not None:
            client = self.get_client_by_name(client)
        elif ':' in route:
            try:
                client_str, route = route.split('/', 1)
                assert ':' in client_str
                host, port = client_str.split(':')
                assert '/' not in host
                port = int(port)
                client = self.get_client_by_sender((host, port))
            except Exception:
                print(f'failed to get client address from OSC route "{route}"')
        else:
            client = next(iter(self.clients.values()))

        if client is None:
            print(f'OSC message failed to send, could not determine client')
            return

        if not route.startswith('/'):
            route = '/'+route
        client.send_message(route, msg)
        if self.verbose > 0:
            print(f"OSC message sent {route}:{msg}")
    
    def handle(self, 
            route:str=None, return_host:str=None, return_port:int=None,
            allow_pos=None, allow_kw=True, lock=True):
        """
        OSC handler decorator supporting mixed args and kwargs, typing.

        The decorated function will receive the OSC route as its first argument.
        Further arguments will be the elements of the OSC message, which can be
        converted in various ways by supplying type annotations (see below).

        If the decorated function returns a value, it should be a tuple beginning
        with the OSC route to reply to, followed by the message contents.

        Args:
            route: OSC path for this handler. If not given,
                use the name of the decorated function.
                If this is a callable, assume the decorator is being used 'bare'
                with all arguments set to default values, i.e. `osc.handle(f)`
                is equivalent to `osc.handle()(f)`.
            return_host: hostname of reply address for return value.
                if not given, reply to sender.
            return_port: port of reply address for return value.
                if not given, reply to sender port.
                NOTE: replying on the same port seems to work with SuperCollider,
                but not with Max.
            kwargs: if True (default), parse OSC message for key-value pairs
                corresponding to named arguments of the decorated function.
            lock: if True (default), use the global iipyper lock around the
                decorated function

        keyword arguments of the decorated function:
            if a string with the same name as a parameter is found, 
            the following item in the OSC message will be used as the value.
            positional arguments can still be used before any keyword pairs,
            like in Python.

            Example:
            ```
            @osc.handle
            def my_func(route, a, b, c=2, d=3):
                print(f'{a=}, {b=}, {c=}, {d=}')
            ```
            OSC message: /my_func/ 0 1 2 'd' 3
            prints: "a=0, b=1, c=2, d=3"

            This is idiomatic in SuperCollider, where flat key, value pairs are
            seen, e.g. `~synth.set('freq', 440, 'amp', 0.2)`
            It's also often easier than managing nested data structures in Max/Pd.

            However, there can be ambiguity when strings are positional arguments.
            It's recommended to avoid using strings as positional arguments, 
            or else to ensure that there will be no collision with the names
            of arguments to the decorated function, 
            for example by adding _ to your argument names.
            you can also disable the keyword argument parsing with `kwargs=False`.

        type annotations:
            object:
                decode a string as JSON
                decode bytes through pickle

            Splat[N]: 
                consume N consecutive items from the OSC message items into a 
                tuple, which forms one argument to the decorated function.

                @osc.handle
                f(a:Splat[2], b:Splat[3]) 
                OSC Message: /f 0 banana 2 3 4.5
                a = (0, banana)
                b = (2, 3, 4.5) 

            Splat[None]: 
                consume items up to the next keyword item into a list
                NOTE: generally should not be used when there are strings 
                in the OSC message, and must be used as a keyword argument, 
                since a str matching an argument name delimits the end of the
                Splat (unless the Splat is the final argument).

                f(a:Splat[None]) -- fine, collects all arguments into a list
                f(a:Splat[None], b:Splat[None]=None) -- breaks if `a` should 
                contain a string `'b'`.
                    OSC message: /f a 0 1 2 b 3 4
                    would call `f('/f', [0,1,2], [3,4])`.

            NDArray: 
                decode a string (either JSON or repr format) to a numpy array
                    JSON format: see `iipyper.ndarray_to_json`
                        (no support for complex dtypes)
                    repr format: see `numpy.array_repr`
                decode bytes to a 1-dimensional float32 array
                    TODO: support annotating dtype via pydantic_numpy types

            other types:
                will be decoded by python-osc and validated by pydantic

        """
        if hasattr(route, '__call__'):
            # bare decorator
            f = route
            route = None
        else:
            f = None

        def decorator(f, route=route, 
                return_host=return_host, return_port=return_port,
                allow_pos=allow_pos, allow_kw=allow_kw):
            # default_route = f'/{f.__name__}/*'
            if route is None:
                route = f'/{f.__name__}'
            # print(route)
            assert isinstance(route, str) and route.startswith('/')

            # get info out of function signature
            sig = inspect.signature(f)

            positional_params = []
            named_params = {}
            has_varp = False
            has_varkw = False
            all_default = True
            pos = True
            for i,(name,p) in enumerate(sig.parameters.items()):
                if i==0:
                    # skip first argument (the OSC route)
                    continue
                # print(p, p.kind)
                
                if p.kind == p.VAR_POSITIONAL:
                    pos = False # no more positional args after variadic
                    all_default = False
                elif p.kind == p.VAR_KEYWORD:
                    pos = False # no more positional args after variadic
                else:
                    if p.default == p.empty:
                        all_default = False
                    named_params[name] = p
                    if pos:
                        positional_params.append(p)
                has_varp |= p.kind==p.VAR_POSITIONAL
                has_varkw |= p.kind==p.VAR_KEYWORD

            # set allow_pos to False if it hasn't been explicitly set to True,
            # and all parameters have defaults or are VAR_KEYWORD
            if allow_pos is None:
                allow_pos = not all_default

            if has_varkw and not allow_kw:
                raise ValueError(f"""
                ERROR: iipyper: OSC handler {f} was created with kwargs=False,
                but the decorated function has a ** argument
                """)
            if has_varp and not allow_pos:
                raise ValueError(f"""
                ERROR: iipyper: OSC handler {f} was created with args=False,
                but the decorated function has a * argument
                """)

            # wrap with pydantic validation decorator
            f = pydantic.validate_call(f)

            def handler(client, address, *osc_items):
                """
                Args:
                    client: (host,port) of sender
                    address: full OSC address
                    *args: content of OSC message
                """
                # print(f'{osc_items=}')
                try:
                    args, kw = _parse_osc_items(
                        osc_items, 
                        (positional_params, named_params, has_varp, has_varkw), 
                        allow_pos, allow_kw,
                        self.verbose
                    )
                except Exception:
                    raise ValueError(f"""
                    {address} {osc_items}
                    failed to parse OSC for function with signature: {sig}
                    """)
                # print(f'{args=}, {kw=}')

                try:
                    r = maybe_lock(f, lock, address, *args, **kw)
                    # if there was a return value,
                    # send it as a message back to the sender
                    if r is not None:
                        if not hasattr(r, '__len__'):
                            print("""
                            value returned from OSC handler should start with route
                            """)
                        else:
                            client = (
                                client[0] if return_host is None else return_host,
                                client[1] if return_port is None else return_port
                            )
                            print('iipyper OSC return', client, r)
                            try:
                                self.get_client_by_sender(client).send_message(
                                    r[0], r[1:])
                            except ValueError:
                                rt = [type(item) for item in r]
                                print(
                                    f'iipyper OSC return to {r[0]} failed,' 
                                    f'possibly an unsupported type in {rt}')
                            
                except pydantic.ValidationError as e:
                    print(f'ERROR: iipyper OSC handler:')
                    for info in e.errors(include_url=False):
                        msg = info['msg']
                        loc = info['loc']
                        inp = info['input']
                        print(f'\t{inp.args} {inp.kwargs}')
                        print(f'\t{msg} {loc}')

            self.add_handler(route, handler)
            return f

        return decorator if f is None else decorator(f)
    
    def args(self, route=None, return_host=None, return_port=None):
        """like `handle`, but only positional arguments are allowed.

        able to parse some edge cases which `handle` cannot.
        """
        return self.handle(
            route, return_host, return_port, 
            allow_pos=True, allow_kw=False)

    def kwargs(self, route=None, return_host=None, return_port=None, 
               json_keys=None):
        """like `handle`, but only keyword arguments are allowed.

        able to parse some edge cases which `handle` cannot.
        """
        if json_keys is not None:
            raise ValueError(f"""
            OSC.kwargs {route}: `json_keys` has been removed.
            replace `json_keys` with type annotation of arguments as `object`.
            consider using `OSC.handle` instead of `OSC.kwargs`.
            """)
        return self.handle(
            route, return_host, return_port,
            allow_pos=False, allow_kw=True)

    def __call__(self, client:str, *a, **kw):
        """
        alternate syntax for `send` with client name first

        `osc('my_client_name', 0, 1, 'message contents', None)`

        Args:
            client: name of OSC client created with `create_client`.
        """
        self.send(*a, client=client, **kw)

class Updater():
    '''
    Rate-limited function call
    '''

    def __init__(self, f, count=30):
        self.f = f
        self.count = count
        self.counter = 0

    def __call__(self):
        self.counter += 1
        if self.counter >= self.count:
            self.f()
            self.counter = 0

class ReceiveUpdater:
    '''
    Decouples event handling from updating
    Updating is rate-limited by a counter
    '''

    def __init__(self, f, state=None, count=5, update=False):
        self.f = f
        self.count = count
        self.counter = 0
        self.update = update
        self.state = state

    def set(self, state):
        '''
        Set the Updater's state
        '''
        self.state = state
        self.update = True

    def __call__(self):
        '''
        Update the target function with internal state
        '''
        self.counter += 1
        if not (self.update and
                self.counter > self.count and
                self.state is not None):
            return
        self.f(*self.state)
        self.counter = 0
        self.update = False

class ReceiveListUpdater:
    '''
    Decouples event handling from updating
    Updating is rate-limited by a counter
    Assumes a list[float] instead of *args
    '''

    def __init__(self, f, state=None, count=5, update=False):
        self.f = f
        self.count = count
        self.counter = 0
        self.update = update
        self.state = state

    def set(self, state):
        '''
        Set the Updater's state
        '''
        self.state = state
        self.update = True

    def __call__(self):
        '''
        Update the target function with internal state
        '''
        self.counter += 1
        if not (self.update and
                self.counter > self.count and
                self.state is not None):
            return
        self.f(self.state)
        self.counter = 0
        self.update = False

class OSCReceiveUpdater(ReceiveUpdater):
    '''
    ReceiveUpdater with an OSC handler
    '''

    def __init__(self, osc, address: str, f, state=None, count=10, update=False):
        super().__init__(f, state, count, update)
        self.osc = osc
        self.address = address
        osc.add_handler(self.address, self.receive)

    def receive(self, address, *args):
        # FIXME: ip:port/args
        '''
        v: first argument to the handler is the IP:port of the sender
        v: or you can use dispatcher.map directly
           and not set needs_reply_address=True
        j: can I get ip:port from osc itself?
        v: if you know the sender ahead of time yeah,
           but that lets you respond to different senders dynamically
        '''
        self.set(args[1:])

class OSCReceiveListUpdater(ReceiveListUpdater):
    '''
    ReceiveListUpdater with an OSC handler
    '''

    def __init__(self, osc, address: str, f, state=None, count=10, update=False):
        super().__init__(f, state, count, update)
        self.osc = osc
        self.address = address
        osc.add_handler(self.address, self.receive)

    def receive(self, address, *args):
        self.set(list(args[1:]))


class OSCSend():
    '''
    Non rate-limited OSC send
    '''
    def __init__(self, osc, address: str, f, count=30, client=None):
        self.osc = osc
        self.address = address
        self.f = f
        self.client = client

    def __call__(self, *args):
        self.osc.send(self.address, *self.f(*args), client=self.client)

class OSCSendUpdater():
    '''
    Rate-limited OSC send
    '''

    def __init__(self, osc, address: str, f, count=30, client=None):
        self.osc = osc
        self.address = address
        self.f = f
        self.count = count
        self.counter = 0
        self.client = client

    def __call__(self):
        self.counter += 1
        if self.counter >= self.count:
            self.osc.send(self.address, *self.f(), client=self.client)
            self.counter = 0

class OSCReceiveUpdaters:
    '''
    o = OSCReceiveUpdaters(osc,
        {"/tolvera/particles/pos": s.osc_set_pos,
         "/tolvera/particles/vel": s.osc_set_vel})
    '''

    def __init__(self, osc, receives=None, count=10):
        self.osc = osc
        self.receives = []
        self.count = count
        if receives is not None:
            self.add_dict(receives, count=self.count)

    def add_dict(self, receives, count=None):
        if count is None:
            count = self.count
        {a: self.add(a, f, count=count) for a, f in receives.items()}

    def add(self, address, function, state=None, count=None, update=False):
        if count is None:
            count = self.count
        self.receives.append(
            OSCReceiveUpdater(self.osc, address, function,
                              state, count, update))

    def __call__(self):
        [r() for r in self.receives]


class OSCSendUpdaters:
    '''
    o = OSCSendUpdaters(osc, client="particles", count=10,
        sends={
            "/tolvera/particles/get/pos/all": s.osc_get_pos_all
        })
    '''

    def __init__(self, osc, sends=None, count=10, client=None):
        self.osc = osc
        self.sends = []
        self.count = count
        self.client = client
        if sends is not None:
            self.add_dict(sends, self.count, self.client)

    def add_dict(self, sends, count=None, client=None):
        if count is None:
            count = self.count
        if client is None:
            client = self.client
        {a: self.add(a, f, count=count, client=client)
                     for a, f in sends.items()}

    def add(self, address, function, state=None, count=None, update=False, client=None):
        if count is None:
            count = self.count
        if client is None:
            client = self.client
        self.sends.append(
            OSCSendUpdater(self.osc, address, function, count, client))

    def __call__(self):
        [s() for s in self.sends]


class OSCUpdaters:
    '''
    o = OSCUpdaters(osc, client="boids", count=10,
        receives={
            "/tolvera/boids/pos": b.osc_set_pos,
            "/tolvera/boids/vel": b.osc_set_vel
        },
        sends={
            "/tolvera/boids/pos/all": b.osc_get_all_pos
        }
    )
    '''

    def __init__(self, osc,
                 sends=None, receives=None,
                 send_count=60, receive_count=10,
                 client=None):
        self.osc = osc
        self.client = client
        self.send_count = send_count
        self.receive_count = receive_count
        self.sends = OSCSendUpdaters(
            self.osc, count=self.send_count, client=self.client)
        self.receives = OSCReceiveUpdaters(self.osc, count=self.receive_count)
        if sends is not None:
            self.add_sends(sends)
        if receives is not None:
            self.add_receives(receives)

    def add_sends(self, sends, count=None, client=None):
        if count is None:
            count = self.send_count
        if client is None:
            client = self.client
        self.sends.add_dict(sends, count, client)

    def add_send(self, send, count=None, client=None):
        if count is None:
            count = self.send_count
        if client is None:
            client = self.client
        self.sends.add(send, client=client, count=count)

    def add_receives(self, receives, count=None):
        if count is None:
            count = self.receive_count
        self.receives.add_dict(receives, count=count)

    def add_receive(self, receive, count=None):
        if count is None:
            count = self.receive_count
        self.receives.add(receive, count=count)

    def __call__(self):
        self.sends()
        self.receives()



# Audio WIP
import sounddevice as sd
# import numpy as np
# import inspect

class Audio:
    instances = [] # 
    def __init__(self, *a, **kw):
        print(sd.query_devices())
        # self.stream = sd.InputStream(*a, **kw) # TODO
        self.stream = sd.Stream(*a, **kw) # TODO
        Audio.instances.append(self)

    # def callback(self, f):
    #     """ 
    #         decorate function as a sounddevice callback.
    #         indata, outdata are [frames, channels] numpy arrays 
    #     """
    #     args = inspect.signature(f).parameters
    #     def cb(
    #         indata: np.ndarray, outdata: np.ndarray, 
    #         frames: int, time: sd.CData, status: sd.CallbackFlags
    #         ):

    #         self.



import time
from contextlib import contextmanager

from .state import _lock

@contextmanager
def profile(label, print=print):
    t = time.perf_counter_ns()
    yield None
    dt = (time.perf_counter_ns() - t)*1e-9
    print(f'{label}:\t {int(1000*dt)} ms')

def maybe_lock(f, lock, *a, **kw):
    if lock:
        with _lock:
            return f(*a, **kw)
    else:
        return f(*a, **kw)

class Lag:
    def __init__(self, coef_up, coef_down=None, val=None):
        self.coef_up = coef_up
        self.coef_down = coef_down or coef_up
        self.val = val

    def __call__(self, val):
        if self.val is None:
            self.val = val
        else:
            coef = self.coef_up if val > self.val else self.coef_down
            self.val = self.val*coef + val*(1-coef)
        return self.val
    
    def hpf(self, val):
        return val - self(val)


from threading import Thread
import time
from numbers import Number

import fire

from .util import *
from .timing import *
from .midi import *
from .osc import *
from .audio import *
from .tui import *
from .state import _lock

_threads = []
def repeat(
        interval:float=None, between_calls:bool=False, 
        lock:bool=True, tick:float=5e-3):
    """
    Decorate a function to be called repeatedly in a loop.
    
    Args:
        interval: time in seconds to repeat at.
            If the decorated function returns a number, 
            use that as the interval until the next call
        between_calls: if True, interval is between call and next call,
            if False, between return and next call
        lock: if True, use the global iipyper lock to make calls thread-safe
        tick: minimum interval to sleep for 
            (will spinlock for the remainder for more precise timing)
    """
    # close the decorator over interval and lock arguments
    def decorator(f):
        def g():
            # clock = Clock(tick)
            while True:
                t = time.perf_counter()
                returned_interval = maybe_lock(f, lock)

                if isinstance(returned_interval, Number):
                    wait_interval = returned_interval
                else:
                    wait_interval = interval

                # replace False or None with 0
                wait_interval = wait_interval or 0

                if between_calls:
                    # interval is between calls to the decorated function
                    elapsed = time.perf_counter() - t
                    wait = wait_interval - elapsed
                else:
                    t = time.perf_counter()
                    wait = wait_interval
                # else interval is between return of one and call and next call
                # print(f'{wait=}')
                # tt = time.perf_counter()
                if wait > 0:
                    sleep = wait - tick
                    if sleep > 0:
                        time.sleep(sleep)
                    spin_end = t + wait_interval
                    while time.perf_counter() < spin_end: pass
                    # print(f'waited = {time.perf_counter() - tt}')
                else:
                    print(
                        f'@repeat function "{f.__name__}" is late by {-wait}')

        th = Thread(target=g, daemon=True)
        th.start()
        _threads.append(th)

    return decorator


_cleanup_fns = []
# decorator to make a function run on KeyBoardInterrupt (before exit)
def cleanup(f=None):
    """Decorate a function to be called when the iipyper app exits."""
    def decorator(f):
        _cleanup_fns.append(f)
        return f

    if f is None: # return a decorator
        return decorator
    else: #bare decorator case; return decorated function
        return decorator(f)

# locking decorator
def lock(f):
    """wrap the decorated function with the global iipyper lock"""
    def decorated(*a, **kw):
        with _lock:
            f(*a, **kw)
    return decorated

def start_audio():
    """start all audio streams"""
    for a in Audio.instances:
        if not a.stream.active:
            a.stream.start()

def run(main=None):
    """call this on your main function to run it as an iipyper app"""
    try:
        if main is not None:
            fire.Fire(main)

        # non-blocking main case:
        start_audio()

        # enter a loop if there is not one in main
        while True:
            time.sleep(3e-2)

    except KeyboardInterrupt:
        # for th in _threads:
            # pass
        for a in Audio.instances:
            a.stream.stop()
            a.stream.close()
        for f in _cleanup_fns:
            f()
        exit(0)


from threading import RLock, Lock

# _lock = Lock()

# using a reentrant lock should allow the user to lock around something
# which may be called from a @repeat, MIDI handler, etc without stalling
_lock = RLock()


"""
`from iipyper.types import *` to get a set of types to use with `OSC.handle`
"""

from typing import NewType, List, Tuple, TypeVar, Iterable, Any, Dict, Optional, Union
from typing_extensions import TypeAliasType
import json

import numpy as np

import pydantic_numpy
from pydantic_numpy import NpNDArray as NDArray

class _Splat(type):
    # cache instances so that e.g. (Splat[None] is Splat[None]) evaluates True
    instances = {}
    @staticmethod
    def __getitem__(n):
        if n in _Splat.instances: return _Splat.instances[n]
        if n is None:
            Splat = NewType('Splat', List)
            r = Splat
        else:
            Splat = NewType('Splat', Tuple[(Any,)*n])
            r = Splat
        _Splat.instances[n] = r
        return r
    
class Splat(metaclass=_Splat):
    """horrible typing crimes to produce annotations for _consume_items
    which pydantic can also validate out of the box.

    `Splat[None]` aliases `List`

    `Splat[2]` aliases `Tuple[Any, Any]`

    `Splat[3]` aliases `Tuple[Any, Any, Any]`
    
    etc
    """

def ndarray_to_json(array: np.ndarray) -> str:
    array_list = array.ravel().tolist()
    array_info = {
        'data': array_list,
        'dtype': str(array.dtype),
        'shape': array.shape
    }
    return json.dumps(array_info)

def ndarray_from_json(json_str: str) -> np.ndarray:
    data = json.loads(json_str)
    array_data = data['data']
    dtype = data.get('dtype', 'float32')
    r = np.array(array_data, dtype=dtype)
    if 'shape' in data:
        r = r.reshape(data['shape'])
    return r

def ndarray_from_repr(repr_str: str) -> np.ndarray:
    from numpy import (array, 
        float16, float32, float64,
        complex64, complex128,
        int8, int16, int32, int64)
    return eval(repr_str)



import sys
try:
    from rich.panel import Panel
    from rich.pretty import Pretty

    from textual.app import App, ComposeResult
    from textual.reactive import reactive
    from textual.widgets import Header, Footer, Static, Button, RichLog, Label
    from textual.css.query import NoMatches, TooManyMatches

    class TUI(App):
        """Base Textual app for iipyper programs."""
        ## e.g.
        # CSS_PATH = 'my_tui.css'
        # BINDINGS = [
        #     ("a", "my_a_action", "Do something when A is pressed"),
        #     ("b", "my_b_action", "Do somethign when B is pressed")]

        def __init__(self):
            super().__init__()
            self.std_log = RichLog(id='std_log')

        def compose(self) -> ComposeResult:
            """Create child widgets for the Textual App.
            
            override this to build the TUI for your iipyper app.
            """
            yield Header()
            yield self.std_log
            yield Footer()

        def set_action(self, f):
            """
            Decorator attaching a function to a Textual Button or key bind.
            
            the name of the decorated function should match the id of a child Button node and/or a key binding in the TUI.
            """
            setattr(self, f'action_{f.__name__}', f)
            return f
        
        def on(self, f):
            """
            Decorator attaching a function to a Textual event

            The name of the decorated function should be an event such as:
                mount
                mouse_move
                click
            """
            if f.__name__=='mount':
                self._mount = f
            else:
                # this doesn't work for on_mount for whatever reason
                setattr(self, f'on_{f.__name__}', f)
            return f
        
        def on_mount(self):
            # self.std_log.write('MOUNT')
            try:
                self._mount()
            except Exception as e:
                print(e)
                pass
        
        def set_mouse_move(self, f):
            self.on_mouse_move = f
            return f

        def on_button_pressed(self, event: Button.Pressed) -> None:
            getattr(self, f'action_{event.button.id}')()


        # def _print(self, k, *v):
        #     for s in (k, *v):
        #         self.std_log.write(s)
        #    # self.std_log.write(' '.join(str(s) for s in (k, *v)))

        def flush(self): pass
        def write(self, s):
            """for redirecting stdout to a file-like"""
            if self.is_running:
                return self.call_from_anywhere(self.std_log.write, s)
            else:
                # TODO: buffer and write after start
                return sys.__stdout__.write(s)

        def print(self, *a, **kw):
            """redirects to the UI's default std output log"""
            kw['file'] = self
            print(*a, **kw)
            # if self.is_running:
            #     self.call_from_anywhere(self._print, *a, **kw)
            #     # try:
            #     #     self.call_from_thread(self._print, *a, **kw)
            #     # except:
            #     #     self._print(*a, **kw)
            # else:
            #     print(*a, **kw)

        def call_from_anywhere(self, f, *a, **kw):
            try:
                return self.call_from_thread(f, *a, **kw)
            except RuntimeError as e:
                return f(*a, **kw)

        def __call__(self, *a, **kw):
            if self.is_running:
                self.call_from_anywhere(self._call, *a, **kw)
                # try:
                #     self.call_from_thread(self._call, *a, **kw)
                # except:
                #     self.do_call(*a, **kw)
            else:
                print(*a)
                for k,v in kw.items():
                    print(k, '->', v)

        def _call(self, **kw):
            """
            by default, expects your TUI to have child nodes
            with a reactive `value` attribute, which updates the node when set.
            then calling my_tui(my_node_id=my_value) will update the node.
            """
            if not self.is_running:
                return
            for k in kw:
                try:
                    node = self.query_one('#'+k)
                    node.value = kw[k]
                except (NoMatches, TooManyMatches):
                    self.print(f'TUI: node "{k}" not found')
                except AttributeError:
                    self.print(f'TUI: node "{k}" lacks value "reactive"')


except ImportError as e:
    print(e.msg)
    print('install package `textual` for terminal user interface')

    class TUI:
        def print(self, *a, **kw):
            print(*a, **kw)
        def __call__(self, *a, **kw):
            pass
        def set_action(self, f):
            pass
        def run(self):
            pass



import functools as ft
import time
from typing import Optional, List, Union

import mido

from .state import _lock

def _get_filter(item):
    if item is None:
        return item
    if (not isinstance(item, str)) and hasattr(item, '__iter__'):
        return set(item)
    return {item}

class MIDI:
    """
    iipyper MIDI object.
    Create one of these and use it to make MIDI handlers and send MIDI:

    ```python
    midi = MIDI()

    @midi.handle
    def my_handler(msg):
        print(msg)

    midi.note_on(channel=0, note=0, velocity=64, time=0)
    ```
    """
    @classmethod
    def print_ports(cls):
        print('Available MIDI inputs:')
        for s in set(mido.get_input_names()):
            print(f'\t{s}')
        print('Available MIDI outputs:')
        for s in set(mido.get_output_names()):
            print(f'\t{s}')
        MIDI.ports_printed = True

    ports_printed = False

    def __init__(self, 
        in_ports:Optional[List[str]]=None, 
        out_ports:Optional[List[str]]=None, 
        virtual_in_ports:int=1, virtual_out_ports:int=1, 
        verbose:int=1, 
        # sleep_time:float=5e-4
        ):
        """
        Args:
            in_ports: list of input devices to open (uses all by default)
            out_ports: list of output devices to open (uses none by default)
            virtual_in_ports: number of 'To iipyper X' ports to create
            virtual_out_ports: number of 'From iipyper X' ports to create
        """
        if not MIDI.ports_printed and verbose:
            MIDI.print_ports()

        self.running = False

        self.verbose = int(verbose)
        # self.sleep_time = sleep_time
        # type -> list[Optional[set[port], Optional[set[channel]], function]
        self.handlers = []

        if isinstance(in_ports, str):
            in_ports = in_ports.split(',')
        if isinstance(out_ports, str):
            out_ports = out_ports.split(',')

        # TODO: fuzzy match port names

        if in_ports is None or len(in_ports)==0:
            in_ports = set(mido.get_input_names())

        self.in_ports = {}  
        for i in range(virtual_in_ports):
            virtual_in = f'To iipyper {i+1}'
            self.in_ports[virtual_in] = mido.open_input(
                virtual_in, virtual=True)
        for port in in_ports:
            try:
                self.in_ports[port] = mido.open_input(
                    port, callback=self.get_callback(port))
            except Exception:
                print(f"""WARNING: MIDI input {port} not found""")

        if self.verbose:
            print(f"""opened MIDI input ports: {list(self.in_ports)}""")

        ##### WIP
        self.out_ports = {}
        for i in range(virtual_out_ports):
            virtual_out = f'From iipyper {i+1}'
            self.out_ports[virtual_out] = mido.open_output(
                virtual_out, virtual=True)

        if out_ports is None:
            out_ports = []
        # if out_ports is None or len(out_ports)==0:
            # out_ports = set(mido.get_output_names())  
        # self.out_ports = {}
        for port in out_ports:
            try:
                self.out_ports[port] = mido.open_output(port)
            except Exception:
                print(f"""WARNING: MIDI output {port} not found""")

        if self.verbose:
            print(f"""opened MIDI output ports: {list(self.out_ports)}""")

        self.start()

    def start(self):
        self.running = True

    def handle(self, *a, **kw):
        """MIDI handler decorator.
        
        Decorated function receives the following arguments:
            `msg`: a [mido](https://mido.readthedocs.io/en/stable/messages/index.html) message
        """
        if len(a):
            # bare decorator
            assert len(a)==1
            assert len(kw)==0
            assert hasattr(a[0], '__call__')
            f = a[0]
            filters = {}
        else:
            # with filter arguments
            for k in kw:
                assert k in {
                    'channel', 'port', 'type', 
                    'note', 'velocity', 'value', 
                    'control', 'program'
                    }, f'unknown MIDI message filter "{k}"'
            filters = {k:_get_filter(v) for k,v in kw.items()}
            f = None

        def decorator(f):
            self.handlers.append((filters, f))
            return f
        
        return decorator if f is None else decorator(f)

    def get_callback(self, port_name):
        # print(port_name)
        def callback(msg):
            if self.verbose > 1:
                print(f'{msg=}')
            if not self.running:
                return
            for filters, f in self.handlers:
                use_handler = (
                    'port' not in filters 
                    or port_name in filters.pop('port'))
                use_handler &= all(
                    filt is None 
                    or not hasattr(msg, k)
                    or getattr(msg, k) in filt
                    for k,filt in filters.items())
                if use_handler:
                    with _lock:
                        print(port_name)
                        f(msg)
        return callback

    def _send_msg(self, port, m):
        """send on a specific port or all output ports"""
        ports = self.out_ports.values() if port is None else [self.out_ports[port]]
        # print(ports)
        for p in ports:
            # print('iipyper send', m)
            # iiuc mido send should already be thread safe
            # with _lock:
            p.send(m)

    # # see https://mido.readthedocs.io/en/latest/message_types.html

    def send(self, m:Union[str,mido.Message], *a, port:Optional[int]=None, **kw):
        """
        send a mido message as MIDI. 
        
        These are equivalent:

        ```python
        midi.send(mido.Message('note_on', channel=0, note=0, velocity=64, time=0))
        midi.send('note_on', channel=0, note=0, velocity=64, time=0)
        midi.note_on(channel=0, note=0, velocity=64, time=0)
        ```

        Args:
            m: a [mido](https://mido.readthedocs.io/en/stable/messages/index.html) message or message type
            port: the MIDI port to send on 
                (or sends on all open ports if not specified)
        """
        # print(f'SEND {time.perf_counter()}')
        if isinstance(m, mido.Message):
            self._send_msg(port, m)
            if len(a)+len(kw) > 0:
                print('warning: extra arguments to MIDI send')
        elif isinstance(m, str):
            try:
                self._send_msg(port, mido.Message(m, *a, **kw))
            except Exception:
                print('MIDI send failed: bad arguments to mido.Message')
                raise
        else:
            print('MIDI send failed: first argument should be a mido.Message or str')

    def __getattr__(self, name):
        if name=='cc': name = 'control_change'
        if name=='pc': name = 'program_change'
        if name in (
            'note_on', 'note_off', 'polytouch', 'control_change', 
            'program_change', 'aftertouch', 'pitchwheel', 'sysex'):
            return lambda *a, **kw: self.send(name, *a, **kw)
        raise AttributeError
        

